"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DPoint = exports.DEGREE_TO_PI = exports.PI_TO_DEGREE = exports.DOUBLE_PI_IN_DEGREE = exports.PI_IN_DEGREE = exports.HALF_PI_IN_DEGREE = exports.EARTH_RADIUS_IN_METERS = exports.WORLD_GEODETIC_SYSTEM = exports.PSEUDO_MERCATOR = void 0;
const DLine_1 = require("./DLine");
const DPolygon_1 = require("./DPolygon");
const diff = 0;
// Meters
exports.PSEUDO_MERCATOR = 'EPSG:3857';
// Degrees
exports.WORLD_GEODETIC_SYSTEM = 'EPSG:4326';
exports.EARTH_RADIUS_IN_METERS = 6371008.8;
const EARTH_IN_MITERS = 20037508.34;
const DEGREES_IN_EARTH = 180;
const MITERS_IN_ONE_DEGREE = EARTH_IN_MITERS / DEGREES_IN_EARTH;
const DEGREES_IN_ONE_MITER = DEGREES_IN_EARTH / EARTH_IN_MITERS;
exports.HALF_PI_IN_DEGREE = 90;
exports.PI_IN_DEGREE = 180;
exports.DOUBLE_PI_IN_DEGREE = 360;
exports.PI_TO_DEGREE = Math.PI / exports.PI_IN_DEGREE;
exports.DEGREE_TO_PI = exports.PI_IN_DEGREE / Math.PI;
class DPoint {
    /**
     *
     * @param x - lng, meters to East, radians to East or width
     * @param y - lat, meters to North, radians to North or height
     * @param z - height
     */
    constructor(x = 0, y = 0, z) {
        this.x = 0;
        this.y = 0;
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        this.properties = {};
        this.x = x;
        this.y = y;
        this.z = z;
    }
    static Zero() {
        return new DPoint(0, 0);
    }
    static parse(c) {
        const { lat, lng } = c;
        if (lat && lng) {
            return new DPoint(lat, lng, 0);
        }
        const [x, y, z] = c;
        return new DPoint(x, y, z);
    }
    static isPoint(p) {
        return p instanceof DPoint;
    }
    static parseFromWKT(wkt) {
        const regexp = /POINT \((?<data>(?:(?!\)).)*?)\)$/miu;
        const data = wkt.trim().toUpperCase();
        const res = regexp.exec(data);
        const [x, y, z] = res.groups.data.split(' ').map(Number);
        return new DPoint(x, y, z);
    }
    static Random() {
        return new DPoint(Math.random(), Math.random());
    }
    getTileFromCoords(zoom = this.z) {
        const x = Math.floor((this.x + exports.PI_IN_DEGREE) / exports.DOUBLE_PI_IN_DEGREE * (Math.pow(2, zoom)));
        const y = Math.floor((1 - Math.log(Math.tan(this.y * exports.PI_TO_DEGREE) + 1 / Math.cos(this.y * exports.PI_TO_DEGREE)) / Math.PI) / 2 * (Math.pow(2, zoom)));
        return new DPoint(x, y, zoom);
    }
    getCoordsFromTile(zoom = this.z) {
        const n = Math.PI - 2 * Math.PI * this.y / (Math.pow(2, zoom));
        const x = this.x / (Math.pow(2, zoom)) * exports.DOUBLE_PI_IN_DEGREE - exports.PI_IN_DEGREE;
        const y = exports.PI_IN_DEGREE / Math.PI * Math.atan((Math.exp(n) - Math.exp(-n)) / 2);
        return new DPoint(x, y, zoom);
    }
    toCoords() {
        if (this.z === undefined) {
            return [this.x, this.y];
        }
        return [this.x, this.y, this.z];
    }
    /**
     * Find line between two points.
     * @param p
     */
    findLine(p) {
        if (this.equal(p)) {
            return this.findLine(p.clone().move(0, 1));
        }
        const a = this.y - p.y - diff;
        const b = p.x - this.x - diff;
        const c = this.x * p.y - p.x * this.y - diff;
        if (a === 0) {
            return new DLine_1.DLine(0, 1, c / b, this, p);
        }
        if (b === 0) {
            return new DLine_1.DLine(1, 0, c / a, this, p);
        }
        return new DLine_1.DLine(a, b, c, this, p);
    }
    findInnerAngle(p1, p3) {
        const a1 = this.findLine(p1).getFi();
        const a2 = this.findLine(p3).getFi();
        if (a2 >= a1) {
            return a2 - a1;
        }
        return a2 + Math.PI * 2 - a1;
    }
    transform(from = exports.PSEUDO_MERCATOR, to = exports.WORLD_GEODETIC_SYSTEM) {
        if (from === to && [exports.PSEUDO_MERCATOR, exports.WORLD_GEODETIC_SYSTEM].includes(from)) {
            return this;
        }
        if (from === exports.PSEUDO_MERCATOR && to === exports.WORLD_GEODETIC_SYSTEM) {
            return this.meters2degrees();
        }
        if (from === exports.WORLD_GEODETIC_SYSTEM && to === exports.PSEUDO_MERCATOR) {
            return this.degrees2meters();
        }
        return this;
    }
    toString() {
        return `${this.x} ${this.y}`;
    }
    getValue() {
        return [this.x, this.y];
    }
    height(z) {
        this.z = z;
        return this;
    }
    toWKT() {
        const { x, y } = this;
        return `POINT (${x} ${y})`;
    }
    distance(p) {
        const dx = p.x - this.x;
        const dy = p.y - this.y;
        return Math.sqrt(dx * dx + dy * dy);
    }
    setX(x) {
        this.x = typeof x === 'number' ? x : x(this);
        return this;
    }
    setY(y) {
        this.y = typeof y === 'number' ? y : y(this);
        return this;
    }
    clone() {
        const p = new DPoint(this.x, this.y, this.z);
        p.properties = Object.assign({}, this.properties);
        return p;
    }
    gt(p) {
        return this.x > p.x && this.y > p.y;
    }
    lt(p) {
        return this.x < p.x && this.y < p.y;
    }
    gtOrEqual(p) {
        return this.gt(p) || this.equal(p);
    }
    ltOrEqual(p) {
        return this.lt(p) || this.equal(p);
    }
    rotate(a) {
        const x = this.x * Math.cos(a) - this.y * Math.sin(a);
        const y = this.x * Math.sin(a) + this.y * Math.cos(a);
        this.x = x;
        this.y = y;
        return this;
    }
    move(x = 0, y = x) {
        let xV = 0;
        let yV = 0;
        if (x instanceof DPoint) {
            xV = this.x + x.x;
            yV = this.y + x.y;
        }
        else {
            xV = this.x + x;
            yV = this.y + y;
        }
        this.x = xV;
        this.y = yV;
        return this;
    }
    asRadians() {
        return this.scale(exports.PI_TO_DEGREE);
    }
    asDegrees() {
        return this.scale(exports.DEGREE_TO_PI);
    }
    round() {
        this.x = Math.round(this.x);
        this.y = Math.round(this.y);
        return this;
    }
    ceil() {
        this.x = Math.ceil(this.x);
        this.y = Math.ceil(this.y);
        return this;
    }
    floor() {
        this.x = Math.floor(this.x);
        this.y = Math.floor(this.y);
        return this;
    }
    toFixed(n = 2) {
        this.x = parseFloat(this.x.toFixed(n));
        this.y = parseFloat(this.y.toFixed(n));
        return this;
    }
    abs() {
        this.x = Math.abs(this.x);
        this.y = Math.abs(this.y);
        return this;
    }
    scale(x = 0, y = x) {
        let xV = 0;
        let yV = 0;
        if (x instanceof DPoint) {
            xV = this.x * x.x;
            yV = this.y * x.y;
        }
        else {
            xV = this.x * x;
            yV = this.y * y;
        }
        this.x = xV;
        this.y = yV;
        return this;
    }
    divide(x = 0, y = x) {
        let xV = 0;
        let yV = 0;
        if (x instanceof DPoint) {
            xV = this.x / x.x;
            yV = this.y / x.y;
        }
        else {
            xV = this.x / x;
            yV = this.y / y;
        }
        this.x = xV;
        this.y = yV;
        return this;
    }
    equal(p) {
        return this.x === p.x && this.y === p.y && this.z === p.z;
    }
    like(p, d = 0.001) {
        const likeX = Math.abs(this.x - p.x) < d;
        const likeY = Math.abs(this.y - p.y) < d;
        const likeZ = this.z === p.z || Math.abs(this.z - p.z) < d;
        return likeX && likeY && likeZ;
    }
    get w() {
        return this.x;
    }
    set w(x) {
        this.x = x;
    }
    get h() {
        return this.y;
    }
    set h(y) {
        this.y = y;
    }
    get area() {
        return this.w * this.h;
    }
    get hip() {
        return Math.sqrt(this.w * this.w + this.h * this.h);
    }
    get min() {
        return Math.min(this.x, this.y);
    }
    get max() {
        return Math.max(this.x, this.y);
    }
    get hipPoint() {
        const { hip } = this;
        return new DPoint(hip, hip);
    }
    get xPoint() {
        const { x } = this;
        return new DPoint(x, x);
    }
    get yPoint() {
        const { y } = this;
        return new DPoint(y, y);
    }
    get wPoint() {
        return this.xPoint;
    }
    get hPoint() {
        return this.yPoint;
    }
    simple(xKey = 'x', yKey = 'y') {
        return {
            [xKey]: this.x,
            [yKey]: this.y
        };
    }
    setIfLessThan(p) {
        this.x = Math.max(this.x, p.x);
        this.y = Math.max(this.y, p.y);
        return this;
    }
    minus() {
        return this.scale(-1);
    }
    orthodromicPath(point, pointsCount = 360) {
        const t = this.clone().asRadians();
        const p = point.clone().asRadians();
        const d = Math.sin(p.x - t.x);
        const step = (p.x - t.x) / (pointsCount - 1);
        return new DPolygon_1.DPolygon(Array.from(new Array(pointsCount))
            .map((v, i) => {
            const x = t.x + step * i;
            const y = Math.atan((Math.tan(t.y) * Math.sin(p.x - x)) / d +
                (Math.tan(p.y) * Math.sin(x - t.x)) / d);
            return new DPoint(x, y).asDegrees();
        }));
    }
    degrees2meters() {
        const x = ((this.x + exports.PI_IN_DEGREE) % exports.DOUBLE_PI_IN_DEGREE - exports.PI_IN_DEGREE) * MITERS_IN_ONE_DEGREE;
        const y = (Math.log(Math.tan(((this.y + exports.HALF_PI_IN_DEGREE) % exports.PI_IN_DEGREE) *
            (Math.PI / exports.DOUBLE_PI_IN_DEGREE))) / exports.PI_TO_DEGREE) * MITERS_IN_ONE_DEGREE;
        this.x = x;
        this.y = y;
        return this;
    }
    meters2degrees() {
        const lon = this.x * DEGREES_IN_ONE_MITER;
        const lat = Math.atan(Math.pow(Math.E, ((this.y / MITERS_IN_ONE_DEGREE) * exports.PI_TO_DEGREE))) *
            (exports.DOUBLE_PI_IN_DEGREE / Math.PI) - exports.HALF_PI_IN_DEGREE;
        this.x = lon;
        this.y = lat;
        return this;
    }
}
exports.DPoint = DPoint;
