"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DPolygon = exports.MIN_POINTS_IN_VALID_POLYGON = void 0;
const DPoint_1 = require("./DPoint");
const DLine_1 = require("./DLine");
const DCircle_1 = require("./DCircle");
const DNumbers_1 = require("./DNumbers");
exports.MIN_POINTS_IN_VALID_POLYGON = 3;
const APPROXIMATION_VALUE = 0.1;
const MAX_CONVEX_ITERATIONS = 100;
const CLOSE_TO_INTERSECTION_DISTANCE = 0.001;
class DPolygon {
    constructor(points = []) {
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        this.properties = {};
        this.holes = [];
        this.pPoints = [];
        this.searchStore = {};
        this.pPoints = points;
    }
    static minAreaRectangleSize(poly) {
        const { first, second, last } = poly.clone().open();
        return new DPoint_1.DPoint(first.distance(second), first.distance(last));
    }
    static toDash(poly) {
        let p = new DPolygon();
        const result = [p];
        let trigger = true;
        for (const point of poly.points) {
            if (trigger) {
                p.push(point.clone());
            }
            if (point.properties.pieceBorder) {
                trigger = !trigger;
                if (trigger) {
                    p = new DPolygon();
                    result.push(p);
                    p.push(point.clone());
                }
            }
        }
        return result;
    }
    static minAreaRectangleDirection(poly) {
        const { first, second, last } = poly.clone().open();
        if (!first || !second || !last) {
            return 0;
        }
        if (first.distance(second) > first.distance(last)) {
            return first.findLine(second).getFi();
        }
        return first.findLine(last).getFi();
    }
    static parseFromWKT(wkt, optProps) {
        const data = wkt.trim().toUpperCase();
        let res = new DPolygon();
        if (data.indexOf('POLYGON') === 0) {
            const regexp = /POLYGON \(\((?<data>(?:(?!\)\)$).)*?)\)\)$/miu;
            const reg = regexp.exec(data);
            const [path, ...holes] = reg.groups.data
                .split('), (')
                .map((p) => new DPolygon(p.split(', ')
                .map((pares) => DPoint_1.DPoint.parse(pares.split(' ').map(Number)))));
            if (holes && holes.length) {
                path.holes = holes;
            }
            res = path;
        }
        if (data.indexOf('LINESTRING') === 0) {
            const regexp = /LINESTRING \((?<data>(?:(?!\)$).)*?)\)$/miu;
            const reg = regexp.exec(data);
            res = new DPolygon(reg.groups.data
                .split(', ').map((t) => DPoint_1.DPoint.parse(t.split(' ').map(Number))));
        }
        if (data.indexOf('POINT') === 0) {
            res = new DPolygon([DPoint_1.DPoint.parseFromWKT(data)]);
        }
        if (optProps) {
            return res.transform(optProps.dataProjection, optProps.featureProjection);
        }
        return res;
    }
    static createSquareBySize(size) {
        return new DPolygon([DPoint_1.DPoint.Zero(), size.clone().setX(0), size.clone(), size.clone().setY(0)]).close();
    }
    set points(p) {
        this.pPoints = p;
    }
    get points() {
        return this.pPoints;
    }
    get maxX() {
        return this.pPoints.reduce((a, r) => Math.max(a, r.x), -Infinity);
    }
    get minX() {
        return this.pPoints.reduce((a, r) => Math.min(a, r.x), Infinity);
    }
    get maxY() {
        return this.pPoints.reduce((a, r) => Math.max(a, r.y), -Infinity);
    }
    get minY() {
        return this.pPoints.reduce((a, r) => Math.min(a, r.y), Infinity);
    }
    get center() {
        return this.leftTop.move(this.size.divide(2));
    }
    get h() {
        return this.maxY - this.minY;
    }
    get w() {
        return this.maxX - this.minX;
    }
    get dY() {
        return this.h;
    }
    get dX() {
        return this.w;
    }
    get extend() {
        const { minX, minY, maxX, maxY } = this;
        return new DPolygon([
            new DPoint_1.DPoint(minX, minY),
            new DPoint_1.DPoint(maxX, minY),
            new DPoint_1.DPoint(maxX, maxY),
            new DPoint_1.DPoint(minX, maxY),
            new DPoint_1.DPoint(minX, minY)
        ]);
    }
    get size() {
        const { w, h } = this;
        return new DPoint_1.DPoint(w, h);
    }
    get leftTop() {
        const { minX, minY } = this;
        return new DPoint_1.DPoint(minX, minY);
    }
    get rightBottom() {
        const { maxX, maxY } = this;
        return new DPoint_1.DPoint(maxX, maxY);
    }
    get length() {
        return this.pPoints.length;
    }
    get fullLength() {
        return this.clone().open().perimeter;
    }
    get perimeter() {
        let p = 0;
        for (let i = 1; i < this.pPoints.length; i++) {
            p += this.pPoints[i - 1].distance(this.pPoints[i]);
        }
        return p;
    }
    get area() {
        const closed = this.deintersection;
        let sum = 0;
        for (let i = 1; i < closed.length; i++) {
            const cur = closed.p(i);
            const prev = closed.p(i - 1);
            sum += prev.x * cur.y - prev.y * cur.x;
        }
        return Math.abs(sum / 2) - this.holes.reduce((a, hole) => a + hole.area, 0);
    }
    get deintersection() {
        const p = this.clone().close();
        for (let i = 0; i < p.length - 1; i++) {
            for (let j = i + 2; j < p.length - 1; j++) {
                const firstLine = p.p(i).findLine(p.p(i + 1));
                const secondLine = p.p(j).findLine(p.p(j + 1));
                const intersectionPoint = firstLine.intersection(secondLine);
                if (intersectionPoint &&
                    ![...firstLine.points, ...secondLine.points].some((t) => t.like(intersectionPoint))) {
                    const part = p.removePart(i, j - i).reverse();
                    p.insertAfter(i, ...part);
                    p.insertAfter(j, intersectionPoint);
                    p.insertAfter(i, intersectionPoint);
                }
            }
        }
        return p;
    }
    get valid() {
        return this.length > exports.MIN_POINTS_IN_VALID_POLYGON;
    }
    get first() {
        return this.p(0);
    }
    get second() {
        return this.p(1);
    }
    get last() {
        return this.p(this.length - 1);
    }
    get minAreaRectangle() {
        const p = this.convex;
        let resultPolygon = new DPolygon();
        let resultArea = Infinity;
        for (let k = 0; k < p.length - 1; k++) {
            const l = p.p(k).findLine(p.p(k + 1));
            let maxWidth = 0;
            let maxWidthPoint1 = null;
            let maxWidthPoint2 = null;
            let maxHeight = 0;
            let maxHeightPoint = null;
            for (let i = 0; i < p.length - 1; i++) {
                const p1 = l.findPoint(l.findPerpendicular(p.p(i)));
                const h = p1.distance(p.p(i));
                if (h >= maxHeight) {
                    maxHeight = h;
                    maxHeightPoint = p.p(i);
                }
                for (let j = i; j < p.length - 1; j++) {
                    const p2 = l.findPoint(l.findPerpendicular(p.p(j)));
                    const w = p1.distance(p2);
                    if (w >= maxWidth) {
                        maxWidth = w;
                        maxWidthPoint1 = p1;
                        maxWidthPoint2 = p2;
                    }
                }
            }
            if (!maxWidthPoint1 || !maxWidthPoint2 || !maxHeightPoint) {
                continue;
            }
            const widthLine = maxWidthPoint1.findLine(maxWidthPoint2);
            const perpendicular1 = widthLine.findPerpendicular(maxWidthPoint1);
            const perpendicular2 = widthLine.findPerpendicular(maxWidthPoint2);
            const tempPolygon = new DPolygon([
                maxWidthPoint1,
                maxWidthPoint2,
                perpendicular2.findPoint(perpendicular2.findPerpendicular(maxHeightPoint)),
                perpendicular1.findPoint(perpendicular1.findPerpendicular(maxHeightPoint))
            ]).close();
            if (tempPolygon.area < resultArea) {
                resultPolygon = tempPolygon;
                resultArea = tempPolygon.area;
            }
        }
        return resultPolygon;
    }
    get convex() {
        let p = this.clone().open();
        const { isClockwise } = p;
        if (!isClockwise) {
            p.reverse();
        }
        let l = 0;
        do {
            const p1 = p.last;
            const p2 = p.first;
            const p3 = p.second;
            const d = p2.findInnerAngle(p1, p3);
            if (d > Math.PI || DNumbers_1.DNumbers.likeZero(DNumbers_1.DNumbers.rad2Deg(d)) || DNumbers_1.DNumbers.likePI(d) || DNumbers_1.DNumbers.like2PI(d)) {
                p.removePart(-1, 1);
            }
            else {
                break;
            }
        } while (p.length);
        p.close();
        let iteration = 0;
        do {
            p = p.deintersection;
            l = p.length;
            for (let i = 1; i < p.length - 1; i++) {
                const p1 = p.p(i - 1);
                const p2 = p.p(i);
                const p3 = p.p(i + 1);
                const d = p2.findInnerAngle(p1, p3);
                if (d > Math.PI || DNumbers_1.DNumbers.likeZero(DNumbers_1.DNumbers.rad2Deg(d)) || DNumbers_1.DNumbers.likePI(d) || DNumbers_1.DNumbers.like2PI(d)) {
                    p.removePart(--i, 1);
                }
            }
            iteration++;
        } while (p.length !== l && iteration < MAX_CONVEX_ITERATIONS);
        if (!isClockwise) {
            p.reverse();
        }
        return p;
    }
    get isClockwise() {
        let sum = 0;
        const p = this.clone().close();
        for (let i = 1; i < p.length; i++) {
            const p1 = p.p(i - 1);
            const p2 = p.p(i);
            sum += (p2.x - p1.x) * (p2.y + p1.y);
        }
        return sum < 0;
    }
    get clockWise() {
        if (this.isClockwise) {
            return this.clone();
        }
        return this.clone().reverse();
    }
    get noHoles() {
        const res = this.clone();
        res.holes = [];
        return res;
    }
    intersection(l) {
        const res = [];
        for (let i = 0; i < this.pPoints.length - 1; i++) {
            const p1 = this.pPoints[i];
            const p2 = this.pPoints[i + 1];
            const line = p1.findLine(p2);
            const intersect = line.intersection(l);
            if (intersect) {
                res.push(intersect);
            }
        }
        return res;
    }
    setCenter(newCenter) {
        return this.clone().move(newCenter.move(this.center.minus()));
    }
    toWKT() {
        let h = '';
        if (this.holes && this.holes.length) {
            h = `, ${this.holes.map((hole) => hole.toString()).join(', ')}`;
        }
        return `POLYGON ((${this.deintersection.pPoints.map((r) => `${r.x} ${r.y}`).join(', ')})${h})`;
    }
    rotate(a) {
        this.pPoints = this.pPoints.map((p) => p.rotate(a));
        this.holes = this.holes.map((h) => h.rotate(a));
        return this;
    }
    filter(f) {
        this.pPoints = this.pPoints.filter(f);
        return this;
    }
    move(x = 0, y) {
        this.pPoints = this.pPoints.map((p) => p.move(x, y));
        this.holes = this.holes.map((h) => h.move(x, y));
        return this;
    }
    scale(x = 0, y) {
        this.pPoints = this.pPoints.map((p) => p.scale(x, y));
        this.holes = this.holes.map((h) => h.scale(x, y));
        return this;
    }
    divide(x = 0, y) {
        this.pPoints = this.pPoints.map((p) => p.divide(x, y));
        this.holes = this.holes.map((h) => h.divide(x, y));
        return this;
    }
    round() {
        this.pPoints = this.pPoints.map((p) => p.round());
        this.holes = this.holes.map((h) => h.round());
        return this;
    }
    floor() {
        this.pPoints = this.pPoints.map((p) => p.floor());
        this.holes = this.holes.map((h) => h.floor());
        return this;
    }
    ceil() {
        this.pPoints = this.pPoints.map((p) => p.ceil());
        this.holes = this.holes.map((h) => h.ceil());
        return this;
    }
    toFixed(n = 2) {
        this.pPoints = this.pPoints.map((p) => p.toFixed(n));
        this.holes = this.holes.map((h) => h.toFixed(n));
        return this;
    }
    map(f) {
        this.pPoints = this.pPoints.map(f);
        this.holes = this.holes.map((h) => h.map(f));
        return this;
    }
    p(index, divide = false) {
        if (divide) {
            let t = index;
            while (t < 0) {
                t += this.length;
            }
            return this.pPoints[t % this.length];
        }
        return this.pPoints[index];
    }
    pop() {
        return this.pPoints.pop();
    }
    push(...args) {
        return this.pPoints.push(...args);
    }
    shift() {
        return this.pPoints.shift();
    }
    unshift(...args) {
        return this.pPoints.unshift(...args);
    }
    reverse() {
        this.pPoints = this.pPoints.reverse();
        this.holes = this.holes.map((h) => h.reverse());
        return this;
    }
    getValue() {
        return (this.pPoints.map((r) => r.getValue()) + this.holes
            .reduce((a, h) => a + h.getValue(), ''));
    }
    transform(from = DPoint_1.PSEUDO_MERCATOR, to = DPoint_1.WORLD_GEODETIC_SYSTEM) {
        this.pPoints = this.pPoints.map((r) => r.transform(from, to));
        this.holes = this.holes.map((h) => h.transform(from, to));
        return this;
    }
    toString() {
        return `(${this.pPoints.map((r) => r.toString()).join(', ')})`;
    }
    close() {
        const p0 = this.first;
        if (p0 && !p0.equal(this.last)) {
            this.push(p0.clone());
        }
        return this;
    }
    open() {
        const p = this.first;
        if (this.length > 2 && p && p.equal(this.last)) {
            this.pop();
        }
        return this;
    }
    height(z) {
        this.map((p) => p.height(z));
        this.holes = this.holes.map((h) => h.height(z));
        return this;
    }
    add(poly) {
        const res = new DPolygon([...this.points, ...poly.points]).close();
        res.holes = [...this.holes, ...poly.holes].map((h) => h.clone());
        return res;
    }
    has(p) {
        return this.pPoints.some((q) => q.equal(p));
    }
    clone() {
        const res = new DPolygon([...this.points.map((r) => r.clone())]);
        res.holes = this.holes.map((h) => h.clone());
        res.properties = this.properties;
        return res;
    }
    equal(p) {
        if (!(p instanceof DPolygon)) {
            return false;
        }
        if (this.clone().open().length !== p.clone().open().length || this.holes.length !== p.holes.length) {
            return false;
        }
        return (this.same(p) &&
            this.holes.reduce((a, hole) => a && p.holes.some((pHoles) => pHoles.same(hole)), true));
    }
    same(p) {
        const pClone = p.clone().open();
        const thisClone = this.clone().open();
        const thisAsString = thisClone.toString();
        return thisClone.points.reduce((a) => {
            const f = pClone.shift();
            pClone.push(f);
            return (a ||
                thisAsString === pClone.toString() ||
                thisAsString ===
                    pClone
                        .clone()
                        .reverse()
                        .toString());
        }, false);
    }
    findIndex(p) {
        return this.points.findIndex((t) => t.equal(p));
    }
    approximation(e = Math.sqrt(this.perimeter) * APPROXIMATION_VALUE) {
        return new DPolygon(this.clone().douglasPeucker(this.pPoints, e));
    }
    insertAfter(index, ...points) {
        this.pPoints.splice(index + 1, 0, ...points);
    }
    removePart(index, count) {
        return this.pPoints.splice(index + 1, count);
    }
    hasSimpleIntersection(p) {
        const extend1 = this.extend;
        const extend2 = p.extend;
        const extend1points = extend1.points;
        const extend2points = extend2.points;
        const in1 = extend1points.some((t) => extend2.simpleInclude(t));
        const in2 = extend2points.some((t) => extend1.simpleInclude(t));
        return in1 || in2;
    }
    simpleInclude(p) {
        return this.simpleIncludeX(p) && this.simpleIncludeY(p);
    }
    drawPolygonOnCanvas(canvas, fillColor, strokeColor, shadowColor, lineWidth = 1, steps = this.length - 1) {
        if (this.length < 2) {
            return;
        }
        const ctx = canvas.getContext('2d');
        if (fillColor) {
            ctx.fillStyle = fillColor;
        }
        if (strokeColor) {
            ctx.strokeStyle = strokeColor;
        }
        if (lineWidth) {
            ctx.lineWidth = lineWidth;
        }
        if (fillColor) {
            ctx.beginPath();
        }
        this.goByPath(ctx, steps % this.length);
        if (shadowColor) {
            ctx.shadowColor = shadowColor;
            ctx.shadowBlur = 0;
            ctx.shadowOffsetX = 1;
            ctx.shadowOffsetY = 1;
        }
        if (fillColor) {
            ctx.closePath();
            ctx.fill();
        }
        if (strokeColor) {
            ctx.stroke();
        }
    }
    clearPolygonOnCanvas(canvas) {
        const ctx = canvas.getContext('2d');
        const old = ctx.globalCompositeOperation;
        ctx.globalCompositeOperation = 'destination-out';
        this.goByPath(ctx);
        ctx.fill();
        ctx.globalCompositeOperation = old;
    }
    contain(p, isBorderInside = false, move = DPoint_1.DPoint.Zero()) {
        const simpleInclude = this.simpleInclude(p);
        if (!simpleInclude) {
            return false;
        }
        const onBorder = this.onBorder(p);
        if (onBorder) {
            return isBorderInside;
        }
        const line = p.findLine(this.leftTop.move(move));
        const poly = this.deintersection;
        const intersectionPoints = [];
        for (let i = 0; i < poly.length - 1; i++) {
            const polygonLine = poly.p(i).findLine(poly.p(i + 1));
            const intersection = line.intersection(polygonLine, CLOSE_TO_INTERSECTION_DISTANCE);
            if (intersection) {
                intersectionPoints.push(intersection);
            }
        }
        const hasCorners = intersectionPoints.some((z) => poly.has(z));
        if (hasCorners) {
            return this.contain2(p, isBorderInside);
        }
        return intersectionPoints.length % 2 === 1;
    }
    onBorder(p) {
        const simpleInclude = this.simpleInclude(p);
        if (simpleInclude) {
            const poly = this.deintersection;
            const hasSamePoint = this.points.some((point) => point.equal(p));
            if (hasSamePoint) {
                return true;
            }
            for (let i = 0; i < poly.length - 1; i++) {
                const p0 = poly.p(i);
                const p1 = poly.p(i + 1);
                const polygonLine = p0.findLine(p1);
                const onBorder = polygonLine.x(p).equal(p) && polygonLine.inRange(p);
                if (onBorder) {
                    return true;
                }
            }
        }
        return false;
    }
    nextStart() {
        this.open();
        this.push(this.shift());
        this.close();
        return this;
    }
    removeDuplicates() {
        for (let i = 0; i < this.length - 1; i++) {
            const p1 = this.p(i);
            const p2 = this.p(i + 1);
            if (p1.equal(p2)) {
                this.removePart(i, 1);
                i--;
            }
        }
        return this;
    }
    static parse(a) {
        return new DPolygon(a.map((r) => DPoint_1.DPoint.parse(r)));
    }
    toArrayOfCoords() {
        return this.pPoints.map((r) => r.toCoords());
    }
    divideToPieces(piecesCount) {
        const { fullLength } = this;
        const pieceLength = fullLength / piecesCount;
        let currentPieceLength = pieceLength;
        for (let i = 0; i < this.pPoints.length - 1; i++) {
            const p1 = this.pPoints[i];
            const p2 = this.pPoints[i + 1];
            if (p1.distance(p2) === currentPieceLength) {
                p2.properties.pieceBorder = true;
                currentPieceLength = pieceLength;
                continue;
            }
            if (p1.distance(p2) - currentPieceLength > 0) {
                const circle = new DCircle_1.DCircle(p1, currentPieceLength);
                const line = p1.findLine(p2);
                const intersectionPoint = line.intersectionWithCircle(circle)
                    .filter((p) => line.inRange(p, CLOSE_TO_INTERSECTION_DISTANCE))[0];
                intersectionPoint.properties.pieceBorder = true;
                this.insertAfter(i, intersectionPoint);
                currentPieceLength = pieceLength;
                continue;
            }
            if (p1.distance(p2) - currentPieceLength < 0) {
                currentPieceLength -= p1.distance(p2);
            }
        }
        return this;
    }
    prepareToFastSearch() {
        this.searchStore = {};
        for (const { x, y, z } of this.points) {
            if (!this.searchStore[x]) {
                this.searchStore[x] = {};
            }
            if (!this.searchStore[x][y]) {
                this.searchStore[x][y] = {};
            }
            this.searchStore[x][y][z || 'undefined'] = true;
        }
    }
    fastHas({ x, y, z }) {
        if (!this.searchStore[x]) {
            return false;
        }
        if (!this.searchStore[x][y]) {
            return false;
        }
        if (!this.searchStore[x][y][z || 'undefined']) {
            return false;
        }
        return this.searchStore[x][y][z || 'undefined'];
    }
    get growingPiecesGenerator() {
        // eslint-disable-next-line @typescript-eslint/no-this-alias,consistent-this
        const polygon = this;
        // eslint-disable-next-line func-names
        return function* () {
            const r = new DPolygon();
            for (const p of polygon.pPoints) {
                r.push(p);
                if (p.properties.pieceBorder) {
                    yield r.clone();
                }
            }
            return r.clone();
        };
    }
    simpleIncludeX(p) {
        const { x } = p;
        return this.minX <= x && this.maxX >= x;
    }
    simpleIncludeY(p) {
        const { y } = p;
        return this.minY <= y && this.maxY >= y;
    }
    douglasPeucker(points, e) {
        let dMax = 0;
        let index = 0;
        const end = points.length - 1;
        const line = points[0].findLine(points[end]);
        for (let i = 1; i < end; i++) {
            const d = line.perpendicularDistance(points[i]);
            if (d > dMax) {
                index = i;
                dMax = d;
            }
        }
        if (dMax >= e) {
            const recResult1 = this.douglasPeucker(points.slice(0, index + 1), e);
            const recResult2 = this.douglasPeucker(points.slice(index), e);
            recResult1.pop();
            return [...recResult1, ...recResult2];
        }
        return [points[0], points[end]];
    }
    goByPath(ctx, steps = this.length - 1) {
        const start = this.first;
        ctx.moveTo(start.x, start.y);
        for (let i = 1; i < (steps % this.length); i++) {
            const { x, y } = this.p(i);
            ctx.lineTo(x, y);
        }
    }
    contain2(p, isBorderInside = false) {
        const simpleInclude = this.simpleInclude(p);
        if (!simpleInclude) {
            return false;
        }
        const onBorder = this.onBorder(p);
        if (onBorder) {
            return isBorderInside;
        }
        const poly = this.deintersection;
        let totalFi = 0;
        for (let i = 0; i < poly.length - 1; i++) {
            const p1 = poly.p(i);
            const p2 = poly.p(i + 1);
            const line1 = new DLine_1.DLine(p1.x - p.x, p1.y - p.y, 0);
            const line2 = new DLine_1.DLine(p2.x - p.x, p2.y - p.y, 0);
            const fiDif = line1.findFi(line2);
            if (line1.vectorProduct(line2).c > 0) {
                totalFi += fiDif;
            }
            else {
                totalFi -= fiDif;
            }
        }
        // eslint-disable-next-line no-magic-numbers
        const eps = Math.PI / 10000;
        let result = false;
        const absTotalFi = Math.abs(totalFi);
        if (absTotalFi < eps) {
            result = false;
        }
        else if (Math.abs(2 * Math.PI - absTotalFi) < eps) {
            result = true;
        }
        else {
            throw new Error('contains2 faild');
        }
        return result;
    }
}
exports.DPolygon = DPolygon;
